@@ -1,23 +1,26 @@
 package JSON::Parse;
 require Exporter;
 @ISA = qw(Exporter);
+
 @EXPORT_OK = qw/parse_json
 		json_to_perl
 		valid_json
+		assert_valid_json
 		validate_json
 		json_file_to_perl/;
+
 %EXPORT_TAGS = (
     all => \@EXPORT_OK,
 );
 use warnings;
 use strict;
 use Carp;
-our $VERSION = '0.24_03';
+our $VERSION = '0.28_01';
 require XSLoader;
 XSLoader::load (__PACKAGE__, $VERSION);
 
-our $true = 1;
-our $false = 0;
+our $true;
+our $false;
 our $null;
 
 sub json_to_perl
@@ -25,6 +28,11 @@ sub json_to_perl
     goto &parse_json;
 }
 
+sub validate_json
+{
+    goto &assert_valid_json;
+}
+
 sub valid_json
 {
     my ($json) = @_;
